﻿#if ROC_GM_PURCHASE
using System;
using System.Collections;
using System.Collections.Generic;
using System.Globalization;
using RocketCoroutine;
using RocketGM;
using RocketGM.Models;
#if !UNITY_EDITOR
using RocketGMPurchase.Models;
#endif
using RocketLog;
using RocketUtils;
using RocketUtils.SerializableDictionary;
using UnityEngine;
using UnityEngine.Purchasing;
using UnityEngine.Purchasing.Security;

namespace RocketGMPurchase
{
    public class RocGMPurchaseController : SingletonBehaviour<RocGMPurchaseController>, IStoreListener
    {
        //[Serializable] private class ProductKeys : SerializableDictionary<string, GMPurchaseProduct> { }

        #region Class Definitions

        public enum GMPurchaseSimulationMode
        {
            None,
            Failed,
            Success
        }

        [Serializable]
        public class LocalProduct
        {
            public ProductDefinition Definition;
            public ProductMetadata Metadata;
            public string Receipt;
            public string TransactionId;
            public string OriginalTransactionId;
            public int IsNewPayment;
            public GMPurchaseProduct GMPurchaseProduct;
            public Product UnityProduct;
            public string Name;
        }

        [Serializable]
        public class GMPurchaseProduct
        {
            public string Id;
            public ProductType Type;
            public string DefaultPrice;
            public string DefaultPriceAndroid;
            public bool IsNoAds;
        }

        [Serializable]
        internal class PurchaseSettings
        {
            public bool IsManualValidation = false;
        }

        #endregion

        private Product[] _unityProductsList;
        private PurchaseFailureReason _purchaseFailureReason;
        private bool _isPurchasing;

        public static event Action<GMPurchaseProduct> PurchaseItemStarted;
        public static event Action<LocalProduct, bool> PurchaseItemSuccessful;

        public static event Action<GMPurchaseProduct, PurchaseFailureReason, bool> PurchaseItemFailed;

        public static event Action RestorationStarted;
        public static event Action RestorationFinished;
        public static event Action ValidationStarted;
        public static event Action ValidationFinished;

        public static event Action<PaymentDTO> OnManualValidationRequired;

        public static event Action StoreProductsListReceived;
        public static event Action<InitializationFailureReason> StoreProductsListReceiveFailed;

        private static IStoreController _mStoreController; // The Unity Purchasing system.
        private static IExtensionProvider _mStoreExtensionProvider; // The store-specific Purchasing subsystems.
        
        private static IGooglePlayConfiguration _mGoogleStoreConfiguration;
        private static IAppleConfiguration _mAppleStoreConfiguration;

        private static IGooglePlayStoreExtensions _mGoogleStoreExtension;
        private static IAppleExtensions _mAppleStoreExtension;

        private static ITransactionHistoryExtensions _transactionHistoryExtensions;
        
        private CrossPlatformValidator _localValidator;

        private Queue<LocalProduct> _waitingPurchaseProductsQueue;
        private LocalProduct _currentProduct;
        private GMPurchaseProduct _tempInappProduct;


        [SerializeField] private List<GMPurchaseProduct> _gmPurchaseProducts;
#if ROC_DEBUG_MODE && UNITY_EDITOR
        [Header("DEBUG")] [SerializeField] private GMPurchaseSimulationMode _simulationMode;
#else
        private GMPurchaseSimulationMode _simulationMode = GMPurchaseSimulationMode.None;
#endif

        public GMPurchaseSimulationMode SimulationMode => _simulationMode;

        [Header("Settings")] [SerializeField] private PurchaseSettings _purchaseSettings;


        private string _productId;
        private string _productName;
        private bool _isValidationInProgress;
        private bool _isRestoreInProgress;
        private bool _userInitiated;

        public bool IsInitialized => _mStoreController != null && _mStoreExtensionProvider != null;

        protected readonly RocLog Log = new RocLog(nameof(RocGMPurchaseController), DebugLevels.Debug);


        private const string ProductNameKey = "gmProductNameKey";

        public string ProcessingProductName
        {
            get => PlayerPrefs.GetString(ProductNameKey, "");
            set => PlayerPrefs.SetString(ProductNameKey, value);
        }


#region Initialization	    		
        protected override void OnAwake()
        {
            DontDestroyOnLoad(gameObject);

            Log.Debug("Initialize started", "OnAwake");
            if (IsInitialized)
            {
                // ... we are done here.
                Log.Info("Already initialized", "OnAwake");
                return;
            }

            InternetConnectionController.DoAfterConnectionAvailable(InitializeBuilder);
        }

        private void InitializeBuilder()
        {
            Log.Debug("Initialize builder started", "InitializeBuilder");
            // Create a builder, first passing in a suite of Unity provided stores.
            var builder = ConfigurationBuilder.Instance(StandardPurchasingModule.Instance());

            foreach (GMPurchaseProduct iapProduct in _gmPurchaseProducts)
            {
                builder.AddProduct(iapProduct.Id, iapProduct.Type);
            }

#if !UNITY_EDITOR
            _localValidator = new CrossPlatformValidator(GooglePlayTangle.Data(), AppleTangle.Data(), Application.identifier);
#endif

            _mGoogleStoreConfiguration = builder.Configure<IGooglePlayConfiguration>();
            _mAppleStoreConfiguration = builder.Configure<IAppleConfiguration>();

#if ROC_GM_REPORTING
            RocGm.DoAfterInitialized(() =>
            {
                _isValidationInProgress = false;
                _waitingPurchaseProductsQueue = new Queue<LocalProduct>();
                UnityPurchasing.Initialize(this, builder);
            });
#endif
        }

        public void OnInitialized(IStoreController controller, IExtensionProvider extensions)
        {
            // Purchasing has succeeded initializing. Collect our Purchasing references.
            Log.Info("OnInitialized: PASS", "OnInitialized");

            // Overall Purchasing system, configured with products for this application.
            _mStoreController = controller;

            // Store specific subsystem, for accessing device-specific store features.
            _mStoreExtensionProvider = extensions;

            _mGoogleStoreExtension = _mStoreExtensionProvider.GetExtension<IGooglePlayStoreExtensions>();
            _mAppleStoreExtension = _mStoreExtensionProvider.GetExtension<IAppleExtensions>();

            _unityProductsList = _mStoreController.products.all;

            foreach (var product in _unityProductsList)
            {
                Log.Debug("Initialized Product ID: " + product.definition.id, "OnInitialized");
            }

            CoroutineController.StartCoroutine(ValidateWaitingPaymentsQueue());

            if (StoreProductsListReceived != null) StoreProductsListReceived();
        }

        public void OnInitializeFailed(InitializationFailureReason error, string message)
        {
            Log.Info($"Billing failed to initialize! The reason was: {message}", "OnInitializeFailed");

            switch (error)
            {
                case InitializationFailureReason.AppNotKnown:
                    Log.Warning("Is your App correctly uploaded on the relevant publisher console?", "OnInitializeFailed");
                    break;
                // Ask the user if billing is disabled in device settings.
                case InitializationFailureReason.PurchasingUnavailable:
                    Log.Warning("Billing disabled!", "OnInitializeFailed");
                    break;
                // Developer configuration error; check product metadata.
                case InitializationFailureReason.NoProductsAvailable:
                    Log.Warning("No products available for purchase!", "OnInitializeFailed");
                    break;
            }

            StoreProductsListReceiveFailed?.Invoke(error);
        }

        private void UpdateApplicationUserName(Product product)
        {
#if UNITY_IOS && !UNITY_EDITOR
            var iso = product.metadata.isoCurrencyCode;

            if (!Enum.TryParse<RocGmCurrencyCode>(iso, out var currencyCode))
            {
                throw new Exception("Currency code is not valid!");
            }

            var currencyCodeString = ((int)currencyCode).ToString();

            if (currencyCodeString.Length > 4)
            {
                currencyCodeString = currencyCodeString.Substring(0, 3);
            }

            currencyCodeString = currencyCodeString.PadLeft(4, '0');

            var price = product.metadata.localizedPrice.ToString(CultureInfo.InvariantCulture);

            if (price.Length > 8)
            {
                price = price.Substring(0, 8);
            }

            price = price.PadLeft(8, '0');
            price = price.Replace('.', 'a');

            var isConsumable = (((int)product.definition.type) + 1).ToString();
            isConsumable = isConsumable.PadLeft(4, '0');

            var format = "{0}-b64a-{3}-{2}-{1:000000000000}"; 
            
            var applicationUsername =
                string.Format($"{format}", price, RocGm.PlayerId, currencyCodeString, isConsumable);

            Log.Info("Application Username : " + applicationUsername, "UpdateApplicationUserName");
            
            _mAppleStoreExtension?.SetApplicationUsername(applicationUsername);
#endif
        }

        private void UpdateObfuscatedId(Product product)
        {
#if UNITY_ANDROID && !UNITY_EDITOR
            var playerId = RocGm.PlayerId.ToString(CultureInfo.InvariantCulture);
            var productType = (((int)product.definition.type) + 1).ToString(CultureInfo.InvariantCulture);

            var obfuscatedId = $"{playerId}.{productType}";
            
            _mGoogleStoreConfiguration?.SetObfuscatedAccountId(obfuscatedId);
            _mGoogleStoreConfiguration?.SetObfuscatedProfileId(obfuscatedId);
#endif
        }

        public void OnInitializeFailed(InitializationFailureReason error)
        {
            // TODO: See if we need to implement this callback.
        }
#endregion

#region Purchase
        public void BuyItem(string productId, string productName = null)
        {
            if (_isPurchasing)
            {
                Log.Warning("Buying Other Product", "BuyItem");
                return;
            }

            _userInitiated = true;

            _productId = productId;
            _productName = productName;

            ProcessingProductName = productName;
            
            var product = GetProductWithId(_productId);

            if (product != null)
            {
                GMPurchaseProduct gmPurchaseProduct = GetGMPurchaseProduct(productId);

                if (gmPurchaseProduct == null)
                {
                    Log.Error(
                        $"Product with id {productId} does NOT exist in GMPurchaseProducts! Please add in prefab!",
                        "BuyItem");
                    return;
                }

                _tempInappProduct = gmPurchaseProduct;

                if (_simulationMode != GMPurchaseSimulationMode.None)
                {
                    SimulatedPurchase(gmPurchaseProduct);
                }
                else
                {
                    BuyProductWithId(gmPurchaseProduct.Id);
                }
            }
            else
            {
                Log.Error(string.Format("Product Type ({0}) does not exist in IapProducts", productId), "BuyItem");
            }
        }

        private void SimulatedPurchase(GMPurchaseProduct product)
        {
            if(PurchaseItemStarted != null)
                PurchaseItemStarted.Invoke(product);

            var boughtProduct = GetProductWithId(product.Id);

            _currentProduct = new LocalProduct
            {
                GMPurchaseProduct = product,
                UnityProduct = boughtProduct,
                Definition = boughtProduct.definition,
                IsNewPayment = _userInitiated ? 1 : 0,
                Metadata = boughtProduct.metadata,
                TransactionId = boughtProduct.transactionID,
                Receipt = boughtProduct.receipt,
                Name = ProcessingProductName
            };

            CoroutineController.DoAfterGivenTime(2f, () =>
            {
                switch (_simulationMode)
                {
                    case GMPurchaseSimulationMode.Success:
                        HandlePurchaseSuccessful();
                        break;

                    case GMPurchaseSimulationMode.Failed:
                        HandlePurchaseFailed(PurchaseFailureReason.PaymentDeclined);
                        break;
                }
            });
        }

        private void BuyProductWithId(string productId)
        {
            if (IsInitialized)
            {
                Product product = GetProductWithId(productId);

                if (product != null && product.availableToPurchase)
                {
                    Log.Debug($"Purchasing product asynchronously: '{product.definition.id}'",
                        "BuyProductWithId");
                    _isPurchasing = true;

                    UpdateApplicationUserName(product);
                    UpdateObfuscatedId(product);

                    _mStoreController.InitiatePurchase(product);

                    if (PurchaseItemStarted != null) PurchaseItemStarted(_tempInappProduct);
                }
                else
                {
                    Log.Warning("FAIL! Not purchasing product, either is not found or is not available for purchase",
                        "BuyProductWithId");
                    _userInitiated = false;
                    OnPurchaseFailed(productId, PurchaseFailureReason.ProductUnavailable);
                }
            }
            else
            {
                Log.Warning("FAIL! Not initialized.", "BuyProductWithId");
                _userInitiated = false;
                OnPurchaseFailed(productId, PurchaseFailureReason.PurchasingUnavailable);
            }
        }

        public PurchaseProcessingResult ProcessPurchase(PurchaseEventArgs args)
        {
            Log.Info("Process Purchase Start", "ProcessPurchase");

            if (!IsInitialized)
            {
                Log.Warning("Purchase Controller not initialized yet, wait for initialization!");

                CoroutineController.DoAfterCondition(() => IsInitialized, () => { ProcessPurchase(args); });
                return PurchaseProcessingResult.Pending;
            }

            _isPurchasing = false;
            Product purchaseProduct = args.purchasedProduct;
            GMPurchaseProduct gmPurchaseProduct = GetGMPurchaseProduct(purchaseProduct.definition.id);

            if (gmPurchaseProduct == null)
            {
                Log.Error("Product does NOT exist in IapProducts. Probably a deleted product trying to restore!",
                    "ProcessPurchase");
                return PurchaseProcessingResult.Complete;
            }

            if (!purchaseProduct.hasReceipt || !purchaseProduct.availableToPurchase)
            {
                Log.Warning(string.Format("!purchaseProduct.hasReceipt {0}, !purchaseProduct.availableToPurchase {1}",
                    !purchaseProduct.hasReceipt, !purchaseProduct.availableToPurchase), "ProcessPurchase");
                OnPurchaseFailed(purchaseProduct.definition.id, PurchaseFailureReason.ProductUnavailable);

                return PurchaseProcessingResult.Complete;
            }

#if UNITY_IOS
            var originalTransactionId = purchaseProduct.appleOriginalTransactionID;
#else
            var originalTransactionId = purchaseProduct.transactionID;
#endif

            LocalProduct localProduct = new LocalProduct
            {
                Definition = purchaseProduct.definition,
                Metadata = purchaseProduct.metadata,
                Receipt = purchaseProduct.receipt,
                TransactionId = purchaseProduct.transactionID,
                OriginalTransactionId = originalTransactionId,
                IsNewPayment = _userInitiated ? 1 : 0,
                GMPurchaseProduct = gmPurchaseProduct,
                UnityProduct = purchaseProduct,
                Name = ProcessingProductName
            };

            Log.Debug($"Product Name: {ProcessingProductName}", "ProcessPurchase");
            Log.Debug("Product Receipt: " + purchaseProduct.receipt, "ProcessPurchase");

            _waitingPurchaseProductsQueue.Enqueue(localProduct);
            Log.Debug("WaitingPurchaseProductsQueue Count >> " + _waitingPurchaseProductsQueue.Count,
                "ProcessPurchase");

            var purchasingResult = _userInitiated ? PurchaseProcessingResult.Pending : PurchaseProcessingResult.Complete;

            return purchasingResult;
        }
#endregion

#region Validation

        IEnumerator ValidateWaitingPaymentsQueue()
        {
            yield return new WaitForSecondsRealtime(15f);
            while (true)
            {
                if (RocGm.Reporting.IsPaymentInProgress || _isValidationInProgress ||
                    _waitingPurchaseProductsQueue.Count <= 0)
                {
                    yield return new WaitForSecondsRealtime(1f);
                    continue;
                }

                ValidateProduct();
            }
        }

        private void ValidateProduct()
        {
            _isValidationInProgress = true;

            Log.Info("Queue count: " + _waitingPurchaseProductsQueue.Count, "ValidateProduct");

            LocalProduct localProduct = _waitingPurchaseProductsQueue.Dequeue();
            _currentProduct = localProduct;

            _productId = localProduct.GMPurchaseProduct.Id;

            Log.Info("InProgress Definition >> " + localProduct.Definition.id, "ValidateProduct");
            Log.Info("InProgress TransactionId >> " + localProduct.TransactionId, "ValidateProduct");
            Log.Info("InProgress Original TransactionId >> " + localProduct.OriginalTransactionId, "ValidateProduct");

            if (ValidationStarted != null) ValidationStarted.Invoke();

#if !UNITY_EDITOR
            try
            {
                Log.Info($"Local validation started.", "ValidateProduct");
                
                var receipts = _localValidator.Validate(localProduct.Receipt);
                
                foreach (var receipt in receipts)
                {
                    Log.Info("Receipt:" +
                             "\nProductId: " + receipt.productID+
                             "\nPurchaseDate: " + receipt.purchaseDate.ToString(CultureInfo.InvariantCulture) +
                             "\nTransactionId: " + receipt.transactionID, "ValidateReceiptProduct");
                        
                    if (StandardPurchasingModule.Instance().appStore == AppStore.GooglePlay)
                    {
                        if (receipt is GooglePlayReceipt googleReceipt)
                        {
                            Log.Info($"Local verification success. Receipt result: \n ProductId : {googleReceipt.productID} \n Transaction Id : {googleReceipt.transactionID}", "ValidateProduct");
                        }
                    }
                    else if (StandardPurchasingModule.Instance().appStore == AppStore.AppleAppStore)
                    {
                        if (receipt is AppleInAppPurchaseReceipt appleReceipt)
                        {
                            Log.Info(
                                $"Local verification success. Receipt result: \n ProductId : {appleReceipt.productID} \n Transaction Id : {appleReceipt.transactionID} \n Original Transaction Id : {appleReceipt.originalTransactionIdentifier}",
                                "ValidateProduct");
                        }
                    }
                }
                
                Log.Info("Local verification validated receipt.", nameof(ValidateProduct));
                
                ValidateWaitingPayments(localProduct);
            }
            catch (IAPSecurityException e)
            {
                Log.Warning("Local verification failed !" + e, nameof(ValidateProduct));
                OnPurchaseFailed(_currentProduct.UnityProduct, PurchaseFailureReason.SignatureInvalid);
            }
#else
            ValidateWaitingPayments(localProduct);
#endif
        }

        private Queue<PaymentDTO> _waitingPaymentValidationQueue = new Queue<PaymentDTO>();

        public bool IsWaitingPaymentValidation => _waitingPaymentValidationQueue.Count > 0;

        private void ValidateWaitingPayments(LocalProduct localProduct)
        {
            var paymentData = new PaymentDTO(_currentProduct.GMPurchaseProduct.Type == ProductType.Subscription,
                localProduct.Receipt,
                _userInitiated,
                _currentProduct.GMPurchaseProduct.Type == ProductType.Consumable,
                _currentProduct.Metadata.localizedPrice,
                _currentProduct.Metadata.isoCurrencyCode,
                localProduct.Name,
                localProduct.OriginalTransactionId);

            if (_purchaseSettings.IsManualValidation)
            {
                if (OnManualValidationRequired != null)
                    OnManualValidationRequired.Invoke(paymentData);
                else
                    _waitingPaymentValidationQueue.Enqueue(paymentData);
                return;
            }

            RocGm.Reporting.Payment(paymentData, 
                OnPurchaseValidationCompleted);
        }

        public void ValidateWaitingPaymentsOnQueue()
        {
            _isValidationInProgress = true;
            while (_waitingPaymentValidationQueue.Count > 0)
            {
                if (OnManualValidationRequired == null) break;

                var paymentData = _waitingPaymentValidationQueue.Dequeue();
                OnManualValidationRequired.Invoke(paymentData);
            }
        }

        // Used externally
        // ReSharper disable once MemberCanBePrivate.Global
        public void OnPurchaseValidationCompleted(bool isSuccess, Dictionary<string, object> result)
        {
            _isValidationInProgress = false;
            _productName = null;
            if (!isSuccess)
            {
                Log.Error("Validation is FAILED. CurrentProduct: " + _currentProduct.TransactionId, "OnPurchaseValidationCompleted");
                OnPurchaseFailed(_currentProduct.Definition.id, PurchaseFailureReason.SignatureInvalid);
                ConfirmPendingPurchase(_currentProduct.UnityProduct);

                _currentProduct = null;
                return;
            }

            OnInAppProductBought(true, _currentProduct.GMPurchaseProduct);
        }

        private void OnInAppProductBought(bool opResult, GMPurchaseProduct gmPurchaseProduct)
        {
            Log.Debug("OpResult: " + opResult, "OnProductBought");
            Log.Debug("Current Product: " + _currentProduct.UnityProduct.transactionID, "OnProductBought");

            ConfirmPendingPurchase(_currentProduct.UnityProduct);

            if (!opResult || gmPurchaseProduct == null)
            {
                OnPurchaseFailed(_currentProduct.UnityProduct, PurchaseFailureReason.Unknown);
                return;
            }

            HandlePurchaseSuccessful();
        }

        private void ConfirmPendingPurchase(Product product)
        {
            Log.Debug(string.Format("product id: {0}", product.definition.id), "ConfirmPendingPurchase");
            Log.Debug(string.Format("Confirm pending purchases. product hasReceipt: {0}", product.hasReceipt), "ConfirmPendingPurchase");
            Log.Debug(string.Format("Confirm pending purchases. initialized: {0}", IsInitialized), "ConfirmPendingPurchase");

            _isValidationInProgress = false;

            if (!IsInitialized || !product.hasReceipt)
                return;

            Log.Debug("Pending purchase condition passed, confirming pending purchase.", "ConfirmPendingPurchase");
            _mStoreController.ConfirmPendingPurchase(product);
        }
#endregion

        private void OnPurchaseFailed(string productId, PurchaseFailureReason failureReason)
        {
            Log.Warning(string.Format("FAIL! Product: '{0}', PurchaseFailureReason: {1}", productId, failureReason), "OnPurchaseFailed");

            HandlePurchaseFailed(failureReason);
        }

        /// <summary>
        /// Unity Callback Function. Should NOT be called explicitly
        /// </summary>
        /// <param name="product"></param>
        /// <param name="failureReason"></param>
        public void OnPurchaseFailed(Product product, PurchaseFailureReason failureReason)
        {
            // Print the failure reason as warning if it is user cancelled.
            if (failureReason == PurchaseFailureReason.UserCancelled)
            {
                Log.Warning(string.Format("FAIL! Product: '{0}', PurchaseFailureReason: {1}", product.definition.storeSpecificId, failureReason), "UnityListenerOnPurchaseFailed");
            }
            else
            {
                Log.Warning(string.Format("FAIL! Product: '{0}', PurchaseFailureReason: {1}", product.definition.storeSpecificId, failureReason), "UnityListenerOnPurchaseFailed");
            }
            HandlePurchaseFailed(failureReason);
        }

        private void HandlePurchaseFailed(PurchaseFailureReason failureReason)
        {
            _isPurchasing = false;

            GMPurchaseProduct gmPurchaseProduct =
                _currentProduct != null ? _currentProduct.GMPurchaseProduct : _tempInappProduct;

            Log.Debug("Waiting Purchase Queue Count: " + _waitingPurchaseProductsQueue?.Count, "HandlePurchaseFailed");
            if (_waitingPurchaseProductsQueue?.Count == 0)
            {
                _isValidationInProgress = false;
                ValidationFinished?.Invoke();
            }

            if (PurchaseItemFailed != null) PurchaseItemFailed(gmPurchaseProduct, failureReason, _userInitiated);

            if (!_isValidationInProgress)
                _userInitiated = false;
        }

        private void HandlePurchaseSuccessful()
        {
            Log.Debug("Waiting Purchase Queue Count: " + _waitingPurchaseProductsQueue.Count,
                "HandlePurchaseSuccessful");
            if (_waitingPurchaseProductsQueue.Count == 0 && ValidationFinished != null)
                ValidationFinished.Invoke();

            if (PurchaseItemSuccessful != null)
            {
                PurchaseItemSuccessful(_currentProduct, _userInitiated);
                if(RocGm.GameSettings.RemotePlayerDataEnabled)
                    RocGm.Reporting.SendRemotePlayerDataWhenPurchaseSuccessful();
            }
            
            _userInitiated = false;
            _currentProduct = null;

            ProcessingProductName = string.Empty;
        }

        public GMPurchaseProduct GetGMPurchaseProduct(string productId)
        {
            foreach (GMPurchaseProduct gmPurchaseProduct in _gmPurchaseProducts)
            {
                if (gmPurchaseProduct.Id == productId)
                    return gmPurchaseProduct;
            }

            return null;
        }

        public string GetProductPrice(string productId)
        {
            var unityProduct = GetProductWithId(productId);

            if (unityProduct != null)
            {
                return unityProduct.metadata.localizedPrice.ToString(CultureInfo.InvariantCulture);
            }

#if UNITY_ANDROID
            if (!string.IsNullOrEmpty(GetGMPurchaseProduct(productId).DefaultPriceAndroid))
                return GetGMPurchaseProduct(productId).DefaultPriceAndroid;
            
            return GetGMPurchaseProduct(productId).DefaultPrice;
#else
            return GetGMPurchaseProduct(productId).DefaultPrice;
#endif
        }
        
        public Product GetProductWithId(string productID)
        {
            if (!IsInitialized)
                return null;

            return _mStoreController.products?.WithID(productID);
        }

        #region Restore
        /// <summary>
        /// USE ONLY FOR IOS!
        /// Restore purchases previously made by this customer. Some platforms automatically restore purchases, like Google.
        /// Apple currently requires explicit purchase restoration for IAP, conditionally displaying a password prompt.
        /// </summary>
        public void RestorePurchases()
        {
            if (!InternetConnectionController.HasConnection)
            {
                return;
            }

            if (_isRestoreInProgress)
            {
                Log.Warning("Already in progress. Ignoring additional request.");
                return;
            }

            // If Purchasing has not yet been set up ...
            if (!IsInitialized)
            {
                // ... report the situation and stop restoring. Consider either waiting longer, or retrying initialization.
                Log.Warning("RestorePurchases: Not initialized. Initialization may still be in progress.");
                return;
            }

            // If we are running on an Apple device ... 
            if (Application.platform == RuntimePlatform.IPhonePlayer ||
                Application.platform == RuntimePlatform.OSXPlayer)
            {
                if (RestorationStarted != null)
                {
                    RestorationStarted();
                    Log.Debug("RestorationStarted event invoked.");
                }

                // ... begin restoring purchases
                Log.Info("RestorePurchases started ...");
                _isRestoreInProgress = true;
                
                // Begin the asynchronous process of restoring purchases. Expect a confirmation response in 
                // the Action<bool> below, and ProcessPurchase if there are previously purchased products to restore.
                _mAppleStoreExtension?.RestoreTransactions((result, error) =>
                {
                    if (!result)
                    {
                        var reason = string.IsNullOrEmpty(error) ? "Unknown error" : error;
                        Log.Error($"RestoreTransactions failed. Aborting restoration. Reason: {reason}");

                        OnTransactionsRestorationCompleted();
                        return;
                    }

                    // Wait for the process purchase
                    //if there is no products to purchase, just close loading
                    CoroutineController.StartCoroutine(WaitForRestoration());
                    // The first phase of restoration. If no more responses are received on ProcessPurchase then 
                    // no purchases are available to be restored.
                    Log.Debug("RestorePurchases continuing: " + result + ". If no further messages, no purchases available to restore.");
                });
            }
            // Otherwise ...
            else
            {
                // We are not running on an Apple device. No work is necessary to restore purchases.
                Log.Error("RestorePurchases: Not supported on this platform. Current = " + Application.platform);
                OnPurchaseFailed(string.Empty, PurchaseFailureReason.ExistingPurchasePending);
            }
        }

        private void OnTransactionsRestorationCompleted()
        {
            _isRestoreInProgress = false;

            if (RestorationFinished != null) RestorationFinished();
        }

        private IEnumerator WaitForRestoration()
        {
            yield return new WaitUntil(() => _waitingPurchaseProductsQueue.Count == 0);
            OnTransactionsRestorationCompleted();
        }

        #endregion

        /// <summary>
        /// Can only be used in DEBUG mode
        /// </summary>
        /// <param name="simulationMode"></param>
        public void ChangeSimulationMode(GMPurchaseSimulationMode simulationMode)
        {
#if ROC_DEBUG_MODE
            _simulationMode = simulationMode;
#endif
        }

    }
}
#endif